/***********************************************************************************************************************
 *
 * blueBill Mobile - open source birdwatching
 * ==========================================
 *
 * Copyright (C) 2009, 2010 by Tidalwave s.a.s. (http://www.tidalwave.it)
 * http://bluebill.tidalwave.it/mobile/
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: TaxonomyFactory.java,v 8233f01770f2 2010/04/24 00:06:31 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.taxonomy;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.text.Collator;
import java.io.BufferedReader;
import java.io.Reader;
import org.json.me.JSONArray;
import org.json.me.JSONException;
import org.json.me.JSONObject;
import it.tidalwave.util.Id;
import it.tidalwave.util.NotFoundException;
import it.tidalwave.bluebill.taxonomy.Taxonomy.Type;
import static it.tidalwave.bluebill.taxonomy.Taxonomy.Type.SPECIES;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: $
 *
 **********************************************************************************************************************/
public class TaxonomyFactory
  {
    private static Map<Type, String> nameMapByType = new HashMap<Type, String>();

    static
      {
        nameMapByType.put(Type.CLASS, "Class");
        nameMapByType.put(Type.ORDER, "Order");
        nameMapByType.put(Type.FAMILY, "Family");
        nameMapByType.put(Type.GENUS, "Genus");
        nameMapByType.put(Type.SPECIES, "Species");
        nameMapByType.put(Type.SUBSPECIES, "Subspecies");
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private final Comparator<Taxon> taxonComparator = new Comparator<Taxon>()
      {
        final Collator collator = Collator.getInstance();

        public int compare (final @Nonnull Taxon taxon1, final @Nonnull Taxon taxon2)
          {
            return collator.compare(taxon1.getScientificName(), taxon2.getScientificName());
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private final Comparator<Taxon> speciesComparator = new Comparator<Taxon>()
      {
        final Collator collator = Collator.getInstance();

        public int compare (final @Nonnull Taxon taxon1, final @Nonnull Taxon taxon2)
          {
            try
              {
                final Taxon parent1 = taxon1.getParent();
                final Taxon parent2 = taxon2.getParent();
                final int r = taxonComparator.compare(parent1, parent2);
                return (r != 0) ? r : collator.compare(taxon1.getScientificName(), taxon2.getScientificName());
              }
            catch (NotFoundException e)
              {
                throw new RuntimeException(e); // never happens for SPECIES
              }
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Taxonomy loadTaxonomy (final @Nonnull Reader r)
      throws Exception
      {
        final BufferedReader br = new BufferedReader(r);
        final StringBuilder builder = new StringBuilder();

        for (;;)
          {
            final String line = br.readLine();

            if (line == null)
              {
                break;
              }

            builder.append(line);
          }

        final String string = builder.toString();
        final JSONObject rootObject = new JSONObject(string);
        final JSONObject jsonTaxonomy = rootObject.getJSONObject("Taxonomy");
        final Taxonomy taxonomy = new Taxonomy(jsonTaxonomy.getString("name"));
        scan(taxonomy, null, jsonTaxonomy, 0);

        final Set<Type> types = new HashSet<Type>(taxonomy.taxonListMapByType.keySet());
        types.remove(Type.SPECIES);
        types.remove(Type.SUBSPECIES);

        for (final Type type : types)
          {
            final List<Taxon> taxons = taxonomy.taxonListMapByType.get(type);
            Collections.sort(taxons, taxonComparator);
          }

        Collections.sort(taxonomy.taxonListMapByType.get(SPECIES), speciesComparator);

        return taxonomy;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    private List<Taxon> scan (final @Nonnull Taxonomy taxonomy,
                              final @Nonnull Taxon parent,
                              final @Nullable JSONObject jsonParent,
                              final @Nonnull int level)
      throws JSONException
      {
        final Type type = Type.values()[level];
        final List<Taxon> children = new ArrayList<Taxon>();
        final String childName = nameMapByType.get(type);
        final JSONArray jsonChildren = jsonParent.optJSONArray(childName);

        if (jsonChildren != null)
          {
            for (int i = 0; i < jsonChildren.length(); i++)
              {
                final JSONObject jsonChild = jsonChildren.getJSONObject(i);
                // FIXME: id should be mandatory, but unfortunately Aves hasn't in EBNItalia 2003
                final Id id = new Id(jsonChild.optString("id"));
                final Taxon child = new Taxon(id, jsonChild.getString("name"), type, parent);
                addDisplayName(jsonChild ,child, Locale.ENGLISH);
                addDisplayName(jsonChild, child, Locale.ITALIAN);

                children.add(child);
                taxonomy.taxonMapById.put(id, child);

                if (level < Type.values().length - 1)
                  {
                    child.children.addAll(scan(taxonomy, child, jsonChild, level + 1));
                  }
              }
          }
        
        List<Taxon> taxons = taxonomy.taxonListMapByType.get(type);
        
        if (taxons == null)
          {
            taxons = children;  
          }
        else
          {
            taxons.addAll(children);  
          }

        taxonomy.taxonListMapByType.put(type, taxons);

        return children;
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private static void addDisplayName (final @Nonnull JSONObject jsonObject,
                                        final @Nonnull Taxon taxon,
                                        final @Nonnull Locale locale)
      {
        final String displayName = jsonObject.optString("lang_" + locale.getLanguage().toLowerCase());

        if (displayName != null)
          {
            taxon.displayNameMap.put(locale, displayName);
          }
      }
  }
